;;; This source file is part of the Meta-CVS program, 
;;; which is distributed under the GNU license.
;;; Copyright 2002 Kaz Kylheku

(require "system")
(require "mapping")
(require "types")
(require "chatter")
(require "options")
(provide "generic")

(defun prepare-file-list (files &key global-if-empty-file-list
                          always-include-version-file)
  "Return a list of mapping-entries and a list of meta files to be
processed and a third value indicating if all files should be
included."
  (when (and *metaonly-option* files)
    (error "cannot specify both --metaonly option and file arguments."))
  (in-sandbox-root-dir
    (let (files-to-process
          (filemap (mapping-read *mcvs-map-local*))
          (global? (and (null files)
                        global-if-empty-file-list
                        (not *metaonly-option*)
                        (not *nometa-option*)))
	  (do-meta-files (and (or *metaonly-option* *meta-option*)
			      (not *nometa-option*))))

      (unless *metaonly-option*
	(chatter-debug "Preparing file list.~%")

        (cond
	  ((and (null files)
		global-if-empty-file-list
		*nometa-option*)
	     (setf files-to-process filemap))
	  ((and (null files)
		(not global-if-empty-file-list))
	     (setf files-to-process 
		   (mapping-prefix-matches filemap
					   (sandbox-translate-path "."))))
	  (files
	     (dolist (file files)
	       (can-restart-here ("Continue preparing file list.")
		 (let* ((full-name (sandbox-translate-path file))
			(abs-name (canonicalize-path
				    (real-to-abstract-path full-name)))
			(entries (mapping-prefix-matches filemap abs-name)))
		   (if (not entries)
		     (error "~a is not known to Meta-CVS." full-name)
		     (setf files-to-process (nconc files-to-process 
						   entries)))))))))

      (values (mapping-extract-kind files-to-process :file)
              (cond ((and do-meta-files (not global?))
                     (current-dir-restore
                      (chdir *mcvs-dir*)
                      (let (metas)
                        (when (exists ".cvsignore")
                          (push ".cvsignore" metas))
                        (when (exists *mcvs-types-name*)
                          (push *mcvs-types-name* metas))
                        (push *mcvs-version-name* metas)
                        (cons *mcvs-map-name* metas))))
                    ((and (not global?) always-include-version-file)
                     (list *mcvs-version-name*))
                    (t
                     nil))
              global?))))

(defun mcvs-generic (cvs-command cvs-options command-options command-args 
		     file-args &key need-sync-before need-sync-after 
		     need-update-after global-if-empty-file-list
                     always-include-version-file
                     no-invoke-cvs)
  (multiple-value-bind (files meta-files global?)
      (prepare-file-list
       file-args
       :global-if-empty-file-list global-if-empty-file-list
       :always-include-version-file always-include-version-file)

    (when (or files meta-files global?)
      (in-sandbox-root-dir
       (when need-sync-before
         (chatter-debug "Synchronizing.~%")
         (mapping-synchronize :filemap files
                              :direction :left))

       (unless no-invoke-cvs
         (current-dir-restore 
          (chdir *mcvs-dir*) 
          (chatter-debug "Invoking CVS.~%")
          (execute-program-xargs `("cvs" ,@(format-opt cvs-options)
                                   ,cvs-command ,@(format-opt command-options)
                                   ,@command-args)
                                 `(,@meta-files
                                   ,@(mapcar #'(lambda (x)
                                                 (basename (mapping-entry-id x)))
                                             files)))))

       (when (and (or meta-files global?) need-update-after)
         (chatter-debug "Updating file structure.~%")
         (mapping-update))
       (when need-sync-after
         (chatter-debug "Synchronizing again.~%")
         (mapping-synchronize :filemap files
                              :direction :right)))))
  (values))

(defun mcvs-diff-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "diff" cvs-options (subst-version-with-tag cvs-command-options)
                nil mcvs-args
                :global-if-empty-file-list t
		:need-sync-before t))

(defun mcvs-tag-wrapper (cvs-options cvs-command-options mcvs-args)
  (if (null mcvs-args)
    (error "specify tag optionally followed by files."))
  (mcvs-generic "tag" cvs-options 
		(subst-version-with-tag cvs-command-options)
                (list (first mcvs-args)) (rest mcvs-args)
		:global-if-empty-file-list t))

(defun mcvs-status-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "status" cvs-options cvs-command-options nil mcvs-args
		:global-if-empty-file-list t
		:need-sync-before t))

(defun mcvs-annotate-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "annotate" cvs-options
                (subst-version-with-tag cvs-command-options)
                nil mcvs-args))

(defun mcvs-watchers-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "watchers" cvs-options cvs-command-options nil mcvs-args))

(defun mcvs-edit-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "edit" cvs-options cvs-command-options nil mcvs-args
		:need-sync-before t))

(defun mcvs-unedit-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "unedit" cvs-options cvs-command-options nil mcvs-args 
		:need-sync-before t
		:need-sync-after t))

(defun mcvs-editors-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "editors" cvs-options cvs-command-options nil mcvs-args))

(defun mcvs-sync-to-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "" cvs-options cvs-command-options nil mcvs-args
		:need-sync-before t
		:no-invoke-cvs t))

(defun mcvs-sync-from-wrapper (cvs-options cvs-command-options mcvs-args)
  (mcvs-generic "" cvs-options cvs-command-options nil mcvs-args
                :global-if-empty-file-list t
		:need-sync-after t
		:no-invoke-cvs t))
