(require "generic")
(provide "log")

(defun raw-version-tree (global-options)
  "Return a hashmap of branch-name to list of (hoc-version
. cvs-version) and another of cvs-version to branch-names."
  (in-sandbox-root-dir
    (chdir *mcvs-dir*)
    (let ((symbols (with-input-from-program (s `("cvs" 
						 ,@(format-opt global-options)
						 "log" "-h"
                                                 ,*mcvs-version-name*))
                     (tags-from-cvs-log s)))
	  (branchpoint-suffix (format nil "~abranch-point" *branch-char*))
          (b (make-hash-table :test 'equal))
          (branch-point-map (make-hash-table :test 'equal)))
      (dolist (s symbols)
	(let* ((tag (first s))
               (file-version (second s))
	       (offset (search branchpoint-suffix tag))
               (v (tag->version tag)))
          (if v
              (push `(,v . ,file-version)
                    (gethash (or (ho-version:version-variant v) "MAIN") b))
              (when (and offset
                         (> offset 0)
                         (= offset
                            (- (length tag) (length branchpoint-suffix))))
                (push (subseq tag 0
                              (- (length tag)
                                 (length branchpoint-suffix)))
                      (gethash file-version branch-point-map))))))
      (values b branch-point-map))))

(defun append-version (prefix v)
  (if prefix
      (ho-version:make-version
       :part-list (append
                   (ho-version:version-part-list prefix)
                   (ho-version:version-part-list v)))
      v))

(defun version-tree (global-options)
  (multiple-value-bind (branches branch-point-map)
      (raw-version-tree global-options)
    (labels ((process-branch (branch branch-point-version)
               (if (endp branch)
                   nil
                   (let* ((v (first branch))
                          (canonical-v (append-version branch-point-version
                                                       (car v))))
                     (make-version-tree-node
                      :version canonical-v
                      :cvs-version (cdr v)
                      :successor (process-branch (rest branch)
                                                 branch-point-version)
                      :branches
                      (mapcar #'(lambda (x)
                                  (process-branch (gethash x branches)
                                                  canonical-v))
                              (gethash (cdr v) branch-point-map)))))))
      (process-branch (gethash "MAIN" branches) nil))))

(defun find-branch (branch tree)
  (if (null tree)
      nil
      (if (equal branch
                 (ho-version:version-variant (version-tree-node-version tree)))
          tree
          (progn
            (dolist (node (version-tree-node-branches tree))
              (when-bind (r (find-branch branch node))
                (return-from find-branch r)))
            (find-branch branch (version-tree-node-successor tree))))))

(defun canonicalize-version (version tree)
  (ho-version:make-version
   :part-list
   (nconc
    (butlast
     (ho-version::version-part-list
      (version-tree-node-version
       (find-branch (ho-version:version-variant version) tree))))
    (last (ho-version::version-part-list version)))))

(defun parse-branch-spec (spec default-branch)
  "branch1,@,branch2,..."
  (let ((branch-names
         (mapcar #'(lambda (name)
                     (cond ((string= name "MAIN")
                            nil)
                           ((string= name "@")
                            default-branch)
                           (t
                            name)))
                 (split-fields spec ","))))
    #'(lambda (x)
        (member (ho-version::version-variant x)
                branch-names
                :test #'equal))))

(defun parse-simple-rev-spec (spec default-version canonicalizer)
  "[!] #v1.2.3 @ #v0.3.0< #v0.3.0<:major :major<=#v0.3.0 @<:patch"
  (labels ((version-string? (s)
             (or (string= s "@")
                 (typep (ignore-errors (read-from-string s))
                        'ho-version:version)))
           (read-difference (s)
             (cond ((or (string= s "") (string= s ":major"))
                    :major)
                   ((string= s ":minor")
                    :minor)
                   ((string= s ":patch")
                    :patch)
                   (t
                    (error "Illegal difference spec: ~A" s))))
           (read-version (s)
             (funcall canonicalizer
                      (cond ((string= s "@")
                             default-version)
                            (t
                             (let ((v (ignore-errors (read-from-string s))))
                               (check-type v ho-version:version)
                               v)))))
           (exact-version-predicate (v)
             #'(lambda (x)
                 (ho-version:version= x v)))
           (less-than-predicate (max diff include-max)
             (if include-max
                 #'(lambda (x)
                     (ho-version:version>= max x diff))
                 #'(lambda (x)
                     (and (not (ho-version:version= x max))
                          (ho-version:version>= max x diff)))))
           (greater-than-predicate (min diff include-min)
             (if include-min
                 #'(lambda (x)
                     (ho-version:version>= x min diff))
                 #'(lambda (x)
                     (and (not (ho-version:version= x min))
                          (ho-version:version>= x min diff)))))
           (parse-xxx (spec)
             (if (not (find #\< spec))
                 (cond ((version-string? spec)
                        (exact-version-predicate (read-version spec)))
                       (t
                        (error "Illegal spec: ~S" spec)))               
                 (let ((include-limit nil))
                   (destructuring-bind (left right)
                       (split-fields spec "<")
                     (when (and (> (length right) 0)
                                (char= (elt right 0) #\=))
                       (setq right (subseq right 1))
                       (setq include-limit t))
                     (cond ((version-string? left)
                            (greater-than-predicate (read-version left)
                                                    (read-difference right)
                                                    include-limit))
                           ((version-string? right)
                            (less-than-predicate (read-version right)
                                                 (read-difference left)
                                                 include-limit))
                           (t
                            (error "Illegal spec: ~S" spec))))))))
    (if (and (> (length spec) 0)
             (char= (elt spec 0) #\!))
        (complement (parse-xxx (subseq spec 1)))
        (parse-xxx spec))))

(defun parse-rev-spec (spec default-version canonicalizer)
  (let ((simple-rev-specs (split-fields spec ","))
        (preds))
    (dolist (s simple-rev-specs)
      (push (parse-simple-rev-spec s default-version canonicalizer)
            preds))
    #'(lambda (v)
        ;; return true iff any of the preds is true
        (some #'(lambda (pred)
                  (funcall pred v))
              preds))))

(defun version-tree-ranges->options (ranges)
  "Return a list of -r options suitable for cvs log"
  (mapcan #'(lambda (x)
              (list
               (list "r"
                     (if (listp x)
                         ;; a real range
                         (format nil "~A:~A"
                                 (version->tag
                                  (version-tree-node-version (car x)))
                                 (version->tag
                                  (version-tree-node-version (cdr x))))
                         ;; a singular version
                         (version->tag
                          (version-tree-node-version x))))))
          ranges))

(defun filter-cvs-log (stream print-tags? print-branches?)
  "Remove some cruft and substitute cvs revisions with hoc module
versions."
  (let ((state :initial)
          version-map
          tags
          branches)
    (labels ((branch-version? (v)
               "See if V denotes a cvs branch version."
               (or (evenp (count #\. v))
                   (string= (second (nreverse (split-fields v ".")))
                            "0")))
             (min-version (version-list)
               "Return the version from VERSION-LIST that is the
ascendant of all others in the list. VERSION-LIST must be empty or
contain exactly one such element."
               ;; quick and dirty: we know that cvs log -h prints tags
               ;; in reverse chronological order and they are pushed
               ;; onto the list, so the first is the min
               (first version-list))
             (reset ()
               (setq version-map (make-hash-table :test 'equal))
               (setq tags ())
               (setq branches ())))
      (reset)
      (do ((line (read-line stream nil) (read-line stream nil)))
          ((null line) nil)
        (ecase state
          ((:initial)
           (when-bind (pos (search "Working file: " line))
             (when (= pos 0)
               (write-line line)))
           (when (string= line "symbolic names:")
             (setq state :syms)))
          ((:syms)
           (cond ((and (not (zerop (length line)))
                       (char= (char line 0) #\tab))
                  (let ((words (split-words line #(#\: #\space #\tab))))
                    (if-bind (v (tag->version (first words)))
                             (push v (gethash (second words) version-map))
                             (if (branch-version? (second words))
                                 (push words branches)
                                 (push words tags)))))
                 (t
                  (setq tags (nreverse tags))
                  (setq branches (nreverse branches))
                  (when print-tags?
                    (write-line "Tags:")
                    (dolist (s tags)
                      (format t "~A~A: ~A~%" #\tab (first s)
                              (or (min-version (gethash (second s) version-map))
                                  (second s)))))
                  (when print-branches?
                    (write-line "Branches:")
                    (dolist (s branches)
                      (format t "~A~A~%" #\tab (first s))))
                  (setq state :description))))
          ((:description)
           (let ((rev-pos (search "revision 1." line)))
             (cond ((= 77 (length line) (count #\= line))
                    (write-line line)
                    (reset)
                    (setq state :initial))
                   ((and rev-pos (= rev-pos 0))
                    (let ((cvs-version (subseq line 9)))
                      (if-bind (v (min-version (gethash cvs-version
                                                        version-map)))
                               (format t "version ~S~%" v)
                               (write-line line))))
                   (t
                    (write-line line))))))))))
  
(defun mcvs-log (global-options command-options command-args)
  (let ((tree (version-tree global-options))
        (preds ())
        (print-tags? t)
        (print-branches? t)
        (print-versions? t)
        (current-version (current-module-version)))
    (multiple-value-bind (files meta-files)
        (prepare-file-list command-args :global-if-empty-file-list t)
      (in-sandbox-root-dir
       (dolist (o command-options)
         (cond ((string= "r" (first o))
                (push (parse-rev-spec (second o)
                                      current-version
                                      #'(lambda (v)
                                          (canonicalize-version v tree)))
                      preds))
               ((string= "b" (first o))
                (push (parse-branch-spec (second o)
                                         (ho-version:version-variant
                                          current-version))
                      preds))
               ((string= "T" (first o))
                (setq print-tags? nil))
               ((string= "B" (first o))
                (setq print-branches? nil))
               ((string= "V" (first o))
                (setq print-versions? nil))))

       (multiple-value-bind (s f)
           (find-versions-in-tree
            #'(lambda (node)
                ;; return true iff all of the preds are true
                (every #'(lambda (pred)
                           (funcall pred (version-tree-node-version node)))
                       preds))
            tree)
         (let ((range-options
                (version-tree-ranges->options
                 (mapcan #'version-tree-ranges
                         (if s
                             (cons s f)
                             f)))))
           (in-sandbox-root-dir
            (chdir *mcvs-dir*)
            (dolist (c (make-commands-xargs
                        `("cvs" ,@(format-opt global-options)
                          "log" ,@(if (or (not range-options)
                                          (not print-versions?))
                                      (list "-h"))
                          ,@(format-opt range-options))
                        `(,@(if files
                                (mapcar #'(lambda (x) 
                                            (basename 
                                             (mapping-entry-id x))) 
                                        files)
                                (list "VERSION"))
                          ,@meta-files)))
              (chatter-debug "invoking ~S in directory ~S"
                             c (namestring (default-directory)))
              (filter-cvs-log (ext:run-program (first c)
                                               :arguments (rest c)
                                               :output :stream)
                              print-tags?
                              print-branches?)))))))))

(defun mcvs-log-wrapper (global-options command-options command-args)
  (mcvs-log global-options command-options command-args))

(defconstant *log-help*
  "Syntax:

  hoc log [ options ] [ file list ]

Options:

  -r version-spec  List only versions defined by version-spec.
  -b branch-list   List only version on branches.
  -T               Do not list tags.
  -B               Do not list branches.
  -V               Do not list versions.

Semantics:

  The log command provides detailed information on the version tree of a
  module or specific files in it.

  A simple invokation without any arguments list all versions, tags
  and branches of the current module.

  By supplying -r and -b parameters it is possible to control which
  versions are listed.

  The branch list following -b is a comma separated list of branch names
  including the special @ symbol that denotes the current branch. This
  option selects versions on any of the listed branches. Examples:

    -b MAIN                  versions on the trunk (branch MAIN)
    -b MAIN,dev              versions on the trunk or branch dev
    -b @                     versions on the current branch

  The -r options is more complex:

    -r #vdev:1.2.3           the given version
    -r @                     the current version
    -r #vdev:1.2.3,#v0.0.0   any of the versions
    -r #vdev:1.2.3<          descendants of the given version
    -r #vdev:1.2.3<:major    descendants of the given version (same as above)
    -r #vdev:1.2.3<:minor    descendants of the given version with no at most
                             minor changes
    -r #vdev:1.2.3<:minor    descendants of the given version with no at most
                             patch changes
    -r #vdev:1.2.3<=         descendants of the given version including itself
    -r !#vdev:1.2.3<         everything except the descendants of #vdev:1.2.3

  Similarly it is also possible to ask for the ascendants:

    -r <#vdev:1.2.3          ascendants of the given version
    -r :patch<#vdev:1.2.3    ascendants of the given version with at most
                             patch changes

  Typical queries:

    hoc log -r '@<:patch'   give me all bug fix commits for the current version

    hoc log -r ':minor<@'   give me all previous versions without crossing major
                            version bumps

  Finally, this complex query lists versions that are:
    - descendants of #v0.0.0 with at most minor revisions and
    - not descendants of #vdev0:10.0.0 and
    - on MAIN or dev0

    hoc log -b MAIN,dev0
            -r '#v0.0.0<=:minor'
            -r '!#vdev0:10.0.0<='")
